body { color:#000000; background-color:#ffffff }
body { font-family: Helvetica, sans-serif; font-size:9pt }
h1 { font-size: 14pt; }
h2 { font-size: 12pt; }
table { font-size:9pt }
table { border-spacing: 0px; border: 1px solid black }
th, table thead {
  background-color:#eee; color:#666666;
  font-weight: bold; cursor: default;
  text-align:center;
  font-weight: bold; font-family: Verdana;
  white-space:nowrap;
}
.W { font-size:0px }
th, td { padding:5px; padding-left:8px; text-align:left }
td.SUMM_DESC { padding-left:12px }
td.DESC { white-space:pre }
td.Q { text-align:right }
td { text-align:left }
tbody.scrollContent { overflow:auto }

table.form_group {
    background-color: #ccc;
    border: 1px solid #333;
    padding: 2px;
}

table.form_inner_group {
    background-color: #ccc;
    border: 1px solid #333;
    padding: 0px;
}

table.form {
    background-color: #999;
    border: 1px solid #333;
    padding: 2px;
}

td.form_label {
    text-align: right;
    vertical-align: top;
}
/* For one line entires */
td.form_clabel {
    text-align: right;
    vertical-align: center;
}
td.form_value {
    text-align: left;
    vertical-align: top;
}
td.form_submit {
    text-align: right;
    vertical-align: top;
}

h1.SubmitFail {
    color: #f00;
}
h1.SubmitOk {
}
